package net.wanji.common.tool.dbtools;

import com.alibaba.druid.pool.DruidDataSource;
import com.google.common.collect.Maps;
import lombok.extern.slf4j.Slf4j;

import java.util.Map;
import java.util.Properties;

/**
 * @author wanji
 */
@Slf4j
public class DSPool {

    private static Map<String, DruidDataSource> dsPools;

    private static final String MYSQL_QUERY_SQL = "SELECT 'X' ";
    private static final String ORACLE_QUERY_SQL = "SELECT 'X' FROM DUAL";
    private static final String MYSQL = "mysql";
    private static final String ORACLE = "oracle";

    static {
        dsPools = Maps.newConcurrentMap();
    }

    public static DruidDataSource getDataSource(String driveName, String url, String userName, String password) {
        String key = getKey(url, userName, password);
        return getDataSource(key, driveName, url, userName, password);
    }

    private static String getKey(String url, String userName, String password) {
        return url + "_" + userName + "_" + password;
    }

    private static DruidDataSource getDataSource(String id, String driveName, String url, String userName, String password) {
        if (!dsPools.containsKey(id)) {
            synchronized (DSPool.class) {
                if (!dsPools.containsKey(id)) {
                    DruidDataSource dataSource = createDataSource(driveName, url, userName, password);
                    if (dataSource != null) {
                        dsPools.put(id, dataSource);
                    }
                }
            }
        }
        return dsPools.get(id);
    }

    public static DruidDataSource createDataSource(String driveName, String url, String userName, String password) {
        DruidDataSource dataSource;
        try {
            dataSource = new DruidDataSource();

            dataSource.setDriverClassName(driveName);
            dataSource.setUrl(url);

            dataSource.setUsername(userName);
            dataSource.setPassword(password);

            // 参数设置
            // 初始化连接数量
            dataSource.setInitialSize(10);
            // 最大连接池数量
            dataSource.setMaxActive(500);
            // 获取连接时最大等待时间
            dataSource.setMaxWait(600 * 1000);

            dataSource.setTestWhileIdle(true);
            dataSource.setTestOnBorrow(false);
            dataSource.setTestOnReturn(false);

            dataSource.setRemoveAbandoned(true);
            //关闭长时间不使用的连接超时时间，单位秒
            dataSource.setRemoveAbandonedTimeout(1800);

            //关闭abanded连接时输出错误日志
            dataSource.setLogAbandoned(true);

            //配置间隔多久才进行一次检测，检测需要关闭的空闲连接，单位是毫秒
            dataSource.setTimeBetweenEvictionRunsMillis(600000);

            // 设置失败尝试次数。
            dataSource.setBreakAfterAcquireFailure(true); // true 中断失败连接
            dataSource.setConnectionErrorRetryAttempts(0); // 失败重试次数

            dataSource.setValidationQueryTimeout(20);
            Properties properties = new Properties();
            properties.setProperty("characterEncoding", "UTF-8");
            // 连接超时时间 毫秒
//            properties.setProperty("spring.datasource.connection-timeout", "10");
            dataSource.setConnectProperties(properties);

            if (driveName.contains(MYSQL)) {
                dataSource.setValidationQuery(MYSQL_QUERY_SQL);
            } else if (driveName.contains(ORACLE)) {
                dataSource.setValidationQuery(ORACLE_QUERY_SQL);
            }
            //timeBetweenEvictionRunsMillis
            //配置间隔多久才进行一次检测，检测需要关闭的空闲连接，单位是毫秒
            dataSource.setTimeBetweenEvictionRunsMillis(10000 * 60 * 3);
            //配置一个连接在池中最小生存的时间，单位是毫秒
            dataSource.setMinEvictableIdleTimeMillis(10000 * 60 * 10);

            dataSource.init();

            log.info("init dataSource pool success,userName: {},password: {},url: {}", userName, password, url);
        } catch (Exception e) {
            log.error("init dataSource pool fail,userName: {},password: {},url: {}", userName, password, url, e);
            dataSource = null;
        }

        return dataSource;
    }

    public static DruidDataSource removeDataSource(String url, String userName, String password) {
        String key = getKey(url, userName, password);
        return removeDataSource(key);
    }

    private static DruidDataSource removeDataSource(String id) {
        return dsPools.remove(id);
    }
}
