package net.wanji.common.tool.dbtools;

import com.alibaba.druid.pool.DruidDataSource;
import net.wanji.common.framework.Constants;
import net.wanji.common.framework.domain.Table;
import net.wanji.common.framework.exception.NodeException;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.dbutils.QueryRunner;
import org.apache.commons.dbutils.handlers.BeanListHandler;
import org.apache.commons.dbutils.handlers.MapListHandler;
import org.apache.commons.dbutils.handlers.ScalarHandler;
import org.apache.commons.lang3.StringUtils;

import java.sql.*;
import java.util.*;

import static net.wanji.common.framework.Constants.DataBase.*;
import static net.wanji.common.framework.Constants.DbDriver.BEYONDB;
import static net.wanji.common.framework.Constants.DbDriver.MYSQL;
import static net.wanji.common.framework.Constants.DbDriver.ORACLE;
import static net.wanji.common.framework.Constants.DbDriver.SQLSERVER;
import static net.wanji.common.framework.Constants.MysqlColumn.BIT;
import static net.wanji.common.framework.Constants.MysqlColumn.FLOAT;
import static net.wanji.common.framework.Constants.MysqlColumn.VARCHAR;
import static net.wanji.common.framework.Constants.MysqlColumn.*;
import static net.wanji.common.framework.Constants.OracleColumn.*;
import static net.wanji.common.framework.Constants.PostGreSqlColumn.DATE;
import static net.wanji.common.framework.Constants.PostGreSqlColumn.TEXT;
import static net.wanji.common.framework.Constants.PostGreSqlColumn.TIMESTAMP;
import static net.wanji.common.framework.Constants.PostGreSqlColumn.*;
import static net.wanji.common.framework.Constants.ServiceGateWayConstants.D2ADataBase.*;
import static net.wanji.common.framework.Constants.TableInfo.*;

/**
 * @author wanji
 */
@Slf4j
public class DSUtils {

    private static final String COLUMN_NAME = "column_name";

    public static void createDatabase(Connection connection, String database) throws SQLException {
        executeUpdateSql(connection, "create database " + database
                + " DEFAULT CHARSET utf8 COLLATE utf8_general_ci");
    }

    public static void deleteDatabase(Connection connection, String database) throws SQLException {
        executeUpdateSql(connection, "drop database " + database);
    }

    /**
     * get the connection
     */
    public static Connection getConn(String driveName, String url, String userName, String password) throws Exception {
        Connection con = null;
        DruidDataSource dataSource = DSPool.getDataSource(driveName, url, userName, password);
        if (dataSource != null) {
            con = dataSource.getConnection();
        }
        return con;
    }

    /**
     * close DB connection
     *
     * @param con DB connection
     */
    public static void closeConn(Connection con) {
        try {
            if (con != null) {
                con.close();
            }
            con = null;
        } catch (Exception ignore) {
            log.error("[DSUtils] 数据库关闭连接失败！");
        }
    }

    /**
     * 判断是否为主键
     *
     * @param conn
     * @param tableName
     * @param columnName
     * @return
     */
    public static boolean isPK(Connection conn, String tableName, String columnName) {
        boolean isPK = false;
        try {
            DatabaseMetaData dsData = conn.getMetaData();
            ResultSet rs = dsData.getPrimaryKeys(null, null, tableName);
            try {
                while (rs.next()) {
                    if (columnName.equals(rs.getString(COLUMN_NAME))) {
                        isPK = true;
                        break;
                    }
                }
            } catch (SQLException e) {
                log.error("query table structure error,tableName:{}, columnName:{}", tableName, columnName, e);
            }
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
        return isPK;
    }

    /**
     * 获取数据库中的所有表名
     *
     * @param conn
     * @return
     * @author cai_cf@126.com
     */
    public static Set<String> getAllTableName(Connection conn) {
        return getAllTableName(conn, null);
    }

    public static Set<String> getAllTableName(Connection conn, String schema) {
        Set<String> tables = new HashSet<>();
        try {
            DatabaseMetaData meta = conn.getMetaData();
            ResultSet rsTables = meta.getTables(conn.getCatalog(), schema,
                    "%", new String[]{"TABLE"});
            while (rsTables.next()) {
                tables.add(rsTables.getString("TABLE_NAME"));
            }
        } catch (SQLException e) {
            log.error("获取数据库中的所有表的表名", e);
        }
        return tables;
    }

    public static List<Table> getAllTableList(Connection conn, String schema) {
        List<Table> tables = new ArrayList<>();
        try {
            DatabaseMetaData meta = conn.getMetaData();
            ResultSet rsTables = meta.getTables(conn.getCatalog(), schema,
                    "%", new String[]{"TABLE"});
            while (rsTables.next()) {
                Table table=new Table();
                table.setNameEn(rsTables.getString("TABLE_NAME"));
                table.setNameCn(rsTables.getString("TABLE_NAME"));
                table.setDes(rsTables.getString("REMARKS"));
                table.setSchemaName(rsTables.getString("TABLE_SCHEM"));
                tables.add(table);
            }
        } catch (SQLException e) {
            log.error("获取数据库中的所有表的表名", e);
        } finally {
            DSUtils.closeConn(conn);
        }
        return tables;
    }

    /**
     * 判断数据库表是否存在
     *
     * @param conn
     * @param schema
     * @param tableName
     * @return
     */
    public static boolean existsTable(Connection conn, String schema, String tableName) {
        boolean flag = false;
        try {
            DatabaseMetaData meta = conn.getMetaData();
            ResultSet rsTables = meta.getTables(null, schema, tableName, null);
            if (rsTables.next()) {
                flag = true;
                log.info(tableName + "已经存在");
            } else {
                log.info(tableName + "不存在");
            }
        } catch (SQLException e) {
            log.error("判断数据库表是否存在错误", e);
        }
        return flag;
    }

    private static String upperCaseTableNameForOracle(Connection conn, String tableName) throws SQLException {
        if (ORACLE.equals(conn.getMetaData().getDatabaseProductName().toLowerCase())) {
            tableName = tableName.toUpperCase();
        }
        return tableName;
    }

    /**
     * 获取表字段属性
     *
     * @param conn
     * @param tableName
     * @return
     * @throws NullPointerException
     */
    public static List<Map<String, String>> findTableColumnList(Connection conn, String tableName) throws NullPointerException {
        return findTableColumnList(conn, null, tableName);
    }

    public static List<Map<String, String>> findTableColumnList(Connection conn, String schema, String tableName) throws NullPointerException {
        verify(tableName, conn);
        List<Map<String, String>> columnsList = Lists.newArrayList();
        try {
            DatabaseMetaData dmd = conn.getMetaData();
            if (dmd == null) {
                log.error("The DatabaseMetaData Object was null.");
                return columnsList;
            }
            ResultSet rs = dmd.getColumns(conn.getCatalog(), schema, tableName, "%");
            rs2List(rs, columnsList, conn, tableName);
        } catch (SQLException e) {
            log.error("数据库连接失败！", e);
        }
        return columnsList;
    }

    /*public static void main(String[] arg) throws Exception {
        String driveName = "oracle.jdbc.driver.OracleDriver";
        String url = "jdbc:oracle:thin:@192.168.5.135:1521:orcl";
        String userName = "AVPLAN";
        String password = "123456";
        Connection conn = getConn(driveName, url, userName, password);
        System.out.println(getTableData(conn, "SELECT * FROM AVPLAN.\"test_\" WHERE \"test_\".ID = 1", new String[]{}));
    }*/

    public static void main(String[] arg) throws Exception {
        String url = "jdbc:mysql://121.36.56.157:63306/uni_bp_dev_db_v1.0.0?useUnicode=true&characterEncoding=UTF-8&connectTimeout=1000";
        String driveName = MYSQL;
        String userName = "root";
        String password = "123456";
        Connection conn = getConn(driveName, url, userName, password);
        System.out.println(getAllTableName(conn));
    }

    private static void verify(String tableName, Connection conn) throws NullPointerException {
        if (StringUtils.isBlank(tableName)) {
            throw new NullPointerException("the table name is null.");
        }
        if (conn == null) {
            throw new NullPointerException("the connection is null");
        }
    }

    private static void rs2List(ResultSet rs, List<Map<String, String>> columnsList, Connection conn, String tableName) {
        try {
            while (rs.next()) {
                Map<String, String> map = Maps.newHashMap();
                String type = rs.getString(TYPE_NAME);
                String isNull = rs.getString(NULLABLE);

                //因JDBC里面强制转换了类型，故特殊处理，by guoliang.dong@wanji.com
                map.put(Constants.TableInfo.COLUMN_NAME, rs.getString(Constants.TableInfo.COLUMN_NAME));
                if("serial".equals(type)){
                    map.put(TYPE_NAME, "int4");
                }else if("bigserial".equals(type)){
                    map.put(TYPE_NAME, "int8");
                }else if("bpchar".equals(type)) {
                    map.put(TYPE_NAME, "char");
                }else {
                    map.put(TYPE_NAME, type);
                }

                map.put(COLUMN_DEF, rs.getString(COLUMN_DEF));

                String dataBaseType = conn.getMetaData().getDatabaseProductName().toLowerCase();
                map.put(COLUMN_SIZE, getColumnSize(dataBaseType, rs));
                map.put(ISNULL, isNull);
                map.put(IS_PK, isPK(conn, tableName, rs.getString(Constants.TableInfo.COLUMN_NAME)) ? "1" : "0");
                columnsList.add(map);
            }
        } catch (Exception e) {
            log.error("query table structure error," + e.getMessage(), e);
        }
    }

    private static String getColumnSize(String dataBaseType, ResultSet rs) {
        String size = null;
        try {
            String type = rs.getString(TYPE_NAME);
            if("serial".equals(type)){
                type = "int4";
            }else if("bigserial".equals(type)){
                type = "int8";
            }else if("bpchar".equals(type)) {
                type = "char";
            }
            //判断数据库类型,来进行判断类型
            switch (dataBaseType) {
                case MYSQL:
                    switch (type) {
                        case YEAR:
                        case DATE:
                        case DATETIME:
                        case BINARY:
                        case VARBINARY:
                        case TINYINT:
                        case SMALLINT:
                        case MEDIUMINT:
                        case INT:
                        case BIGINT:
                        case BIT:
                        case CHAR:
                        case VARCHAR:
                        case Constants.MysqlColumn.TEXT:
                        case LONGTEXT:
                            size = rs.getString(COLUMN_SIZE);
                            break;
                        case DECIMAL:
                        case DOUBLE:
                        case FLOAT:
                            size = rs.getString(COLUMN_SIZE) + "," + rs.getString(DECIMAL_DIGITS);
                            break;
                        default:
                            break;
                    }
                    break;
                case POSTGRESQL:
                    switch (type) {
                        case FLOAT4:
//                            size = rs.getFloat(COLUMN_SIZE)+"";
                            size = "24";
                            break;
                        case FLOAT8:
//                            size = rs.getFloat(COLUMN_SIZE)+"";
                            size = "53";
                            break;
                        case MONEY:
                            size = rs.getString(COLUMN_SIZE);
                            break;
                        case TEXT:
                            break;
                        case DATE:
                            size = rs.getString(COLUMN_SIZE);
                            break;
                        case BPCHAR:
                        case Constants.PostGreSqlColumn.VARCHAR:
                            size = rs.getString(COLUMN_SIZE);
                            break;
                        case INT4:
//                            size = rs.getInt(COLUMN_SIZE)+"";
                            size = "32";
                            break;
                        case INT2:
//                            size = rs.getInt(COLUMN_SIZE)+"";
                            size = "16";
                            break;
                        case Constants.PostGreSqlColumn.BIT:
                        case NUMERIC:
                            size = rs.getString(COLUMN_SIZE);
                            if(size.equals("131089")){
                                size = "0";
                            }
                            break;
                        case BOOL:
                        case INT8:
//                            size = rs.getInt(COLUMN_SIZE)+"";
                            size = "64";
                            break;
                        case TIMESTAMP:
                            size = rs.getString(COLUMN_SIZE);
                            break;
                        case TIME:
                        case BYTEA:
                        case BOX:
                        case CIDR:
                        case INET:
                        case MACADDR:
                        case VARBIT:
                        case CIRCLE:
                        case INTERVAL:
                        case LINE:
                        case LSEG:
                        case POLYGON:
                        case POINT:
                        case PATH:
                            size = rs.getString(COLUMN_SIZE);
                            break;
                        default:
                            break;
                    }
                    break;
                case ORACLE:
                    switch (type) {
                        case VARCHAR2:
                        case ORACLE_CHAR:
                        case RAW:
                        case NVARCHAR2:
                        case Constants.OracleColumn.TIMESTAMP:
                        case TIMESTAMP_WITH_LOCAL_TIME_ZONE:
                        case TIMESTAMP_WITH_TIME_ZONE:
                            size = rs.getString(COLUMN_SIZE);
                            break;

                        case INTERVAL_DAY_TO_SECOND:
                        case INTERVAL_YEAR_TO_MONTH:
                        case NUMBER:
                            size = "0".equals(rs.getString(DECIMAL_DIGITS))
                                    ? rs.getString(COLUMN_SIZE)
                                    : rs.getString(COLUMN_SIZE) + "," + rs.getString(DECIMAL_DIGITS);
                            break;
                        default:
                            break;
                    }
                    break;
                default:
                    throw new NodeException("not support type ");
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return size;
    }

    public static List<Map<String, String>> findGPTableColumnList(Connection conn, String tableName) {
        verify(tableName, conn);
        List<Map<String, String>> columnsList = Lists.newArrayList();
        try {
            DatabaseMetaData dmd = conn.getMetaData();
            if (dmd == null) {
                log.error("The DatabaseMetaData Object was null .");
                return columnsList;
            }

            ResultSet rs = dmd.getColumns(null, null, tableName, "%");
            try {
                while (rs.next()) {
                    Map<String, String> map = Maps.newHashMap();
                    String type = rs.getString(TYPE_NAME);
                    String name = rs.getString(Constants.TableInfo.COLUMN_NAME);
                    if ("UNKNOWN".equals(type)) {
                        type = getColumnType(tableName, name, conn);
                    }
                    String isNull = rs.getString(NULLABLE);

                    map.put(Constants.TableInfo.COLUMN_NAME, name);
                    map.put(TYPE_NAME, type);
                    map.put(COLUMN_DEF, rs.getString(COLUMN_DEF));
                    assert type != null;
                    if (!type.toLowerCase().contains("date") || !type.toLowerCase().contains("time")) {
                        map.put(COLUMN_SIZE, rs.getString(COLUMN_SIZE));
                    }
                    map.put(ISNULL, isNull);
                    map.put(IS_PK, isPK(conn, tableName, rs.getString(Constants.TableInfo.COLUMN_NAME)) ? "1" : "0");
                    columnsList.add(map);
                }
            } catch (Exception e) {
                log.error("query table structure error," + e.getMessage(), e);
            }
        } catch (SQLException e) {
            log.error("数据库连接失败！", e);
        }
        return columnsList;
    }

    private static String getColumnType(String table, String col, Connection con) throws SQLException {
        ResultSet rs = null;
        PreparedStatement pstmt = null;
        String sql = "SELECT t.typname FROM pg_attribute a,pg_class c,pg_type t where a.attrelid=c.oid and t.oid=a.atttypid and c.relname=? and a.attname=?;";
        try {
            pstmt = con.prepareStatement(sql);
            pstmt.setString(1, table);
            pstmt.setString(2, col);
            rs = pstmt.executeQuery();
            if (rs.next()) {
                return rs.getString(1);
            }
        } finally {
            if (rs != null) {
                try {
                    rs.close();
                } catch (Exception e) {
                    rs = null;
                }
            }
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (Exception e) {
                    pstmt = null;
                }
            }
        }
        return null;
    }

    /**
     * 获取总条数
     *
     * @param conn
     * @param sql
     * @return
     * @throws SQLException
     */
    public static int getCount(Connection conn, String sql) throws SQLException {
        int count;
        try {
            QueryRunner queryRunner = new QueryRunner();
            count = Integer.valueOf(queryRunner.query(conn, sql, new ScalarHandler(1)).toString());
        } catch (SQLException e) {
            log.error("Count data of db fail.{sql:" + sql + "}", e);
            throw e;
        }
        return count;
    }

    public static int getCount(Connection conn, String sql, Object[] params) throws SQLException {
        int count;
        try {
            QueryRunner queryRunner = new QueryRunner();
            count = Integer.valueOf(queryRunner.query(conn, sql, new ScalarHandler(1), params).toString());
        } catch (SQLException e) {
            log.error("Count data of db fail.{sql:" + sql + "}", e);
            throw e;
        }
        return count;
    }

    /**
     * execute sql to db (Create Table && Drop Table)
     *
     * @param conn connection of db
     * @param sql  sql to execute
     * @return execute success?
     * @author cai_cf@126.com
     */
    public static boolean executeUpdateSql(Connection conn, String sql) throws SQLException {
        boolean flag = true;
        try {
            conn.setAutoCommit(false);
            Statement stmt = conn.createStatement();
            stmt.executeUpdate(sql);
            // d2a外部事务实现，如果开启，将数据库连接自动提交关闭
            if (!Constants.OPEN_GLOBAL_TRANSACTION.get()) {
                conn.commit();
            }
        } catch (SQLException e) {
            flag = false;
            try {
                conn.rollback();
            } catch (SQLException e1) {
                log.error("executeUpdateSql rollBack error", e1);
                throw e1;
            }
            log.error("executeUpdateSql error sql:{}", sql, e);
            throw e;
        }
        return flag;
    }

    /**
     * execute sql to db (Create Table && Drop Table)
     *
     * @param conn connection of db
     * @param sql  sql to execute
     * @return execute success?
     * @author cai_cf@126.com
     */
    public static int executeSqlReturnChangedRows(Connection conn, String sql) {
        int result = 0;
        try {
            conn.setAutoCommit(false);

            Statement stmt = conn.createStatement();
            result = stmt.executeUpdate(sql);
            // d2a外部事务实现，如果开启，将数据库连接自动提交关闭
            if (!Constants.OPEN_GLOBAL_TRANSACTION.get()) {
                conn.commit();
            }
        } catch (SQLException e) {
            try {
                conn.rollback();
            } catch (SQLException e1) {
                log.error("executeSqlReturnChangedRows rollBack error.", e1);
            }
            log.error("executeSqlReturnChangedRows error.{ sql:{}", sql, e);
        }
        return result;
    }

    public static int executeSqlReturnChangedRows(Connection conn, String sql,Object[] params) {
        int result = 0;
        QueryRunner queryRunner = new QueryRunner();

        try {
            conn.setAutoCommit(false);

            //Statement stmt = conn.createStatement();
            //result = stmt.executeUpdate(sql);
            queryRunner.update(conn, sql, params);
            // d2a外部事务实现，如果开启，将数据库连接自动提交关闭
            if (!Constants.OPEN_GLOBAL_TRANSACTION.get()) {
                conn.commit();
            }
        } catch (SQLException e) {
            try {
                conn.rollback();
            } catch (SQLException e1) {
                log.error("executeSqlReturnChangedRows rollBack error.", e1);
            }
            log.error("executeSqlReturnChangedRows error.{ sql:{}", sql, e);
        }
        return result;
    }

    /**
     * 查询数据
     *
     * @param conn
     * @param sql
     * @param params
     * @return
     * @throws SQLException
     */
    public static List<Map<String, Object>> search(Connection conn, String sql, List<Object> params) throws SQLException {
        List<Map<String, Object>> result;
        QueryRunner runner = new QueryRunner();
        try {
            if (params != null && params.size() > 0) {
                Object[] objects = params.toArray();
                result = runner.query(conn, sql, new MapListHandler(), objects);
            } else {
                result = runner.query(conn, sql, new MapListHandler());
            }
        } catch (SQLException e) {
            log.error("数据查询失败！", e);
            throw e;
        }
        return result;
    }

    public static String searchOne(Connection conn, String sql) throws SQLException {
        String result;
        QueryRunner runner = new QueryRunner();
        try {
            result = runner.query(conn, sql, new ScalarHandler<String>());
        } catch (SQLException e) {
            log.error("数据查询失败！", e);
            throw e;
        }
        return result;
    }

    /**
     * 查所有数据   clob  和Blob需要单独处理
     */
    public static List<Map<String, Object>> getTableData(Connection conn, String sql) throws SQLException {
        List<Map<String, Object>> result;
        QueryRunner runner = new QueryRunner();
        try {
            result = runner.query(conn, sql, new MapListHandler());
        } catch (SQLException e) {
            log.error("getTableData error", e);
            throw e;
        }
        return result;
    }

    public static Map<String, Object> getTableDataObj(Connection conn, String sql) {
        Map<String, Object> map = null;
        QueryRunner runner = new QueryRunner();
        try {
            List<Map<String, Object>>  result = runner.query(conn, sql, new MapListHandler());
            if(result != null && result.size()>0){
                map = result.get(0);
            }
        } catch (SQLException e) {
            log.error("获取一条数据异常", e);
        }
        return map;
    }

    /**
     * 获取表主键
     *
     * @param conn
     * @param tableName
     * @return
     */
    public static String getPrimaryKey(Connection conn, String tableName) throws SQLException, ClassNotFoundException {
        DatabaseMetaData metaData = conn.getMetaData();
        ResultSet resultSet = metaData.getPrimaryKeys(conn.getCatalog(), null, tableName);
        String pkName = "";
        while (resultSet.next()) {
            pkName = resultSet.getString("COLUMN_NAME");
        }
        return pkName;
    }

    /**
     * 查询表中所有数据。
     * <p>
     * 使用Apache DbUtils中的QueryRunner。不会发生sql注入。
     * 与search方法功能类似。
     *
     * @param conn 数据源连接。
     * @param sql  select sql语句。
     * @return 查询结果。
     * @throws SQLException 如果查询失败。
     */
    public static List<Map<String, Object>> getTableData(Connection conn, String sql, Object[] params) throws SQLException {
        List<Map<String, Object>> result;
        QueryRunner runner = new QueryRunner();
        try {
            result = runner.query(conn, sql, new MapListHandler(), params);
        } catch (SQLException e) {
            throw new SQLException("[DSUtils] getTableData error. sql:{}", sql, e);
        }
        return result;
    }

    /**
     * 执行update语句。
     * <p>
     * 此方法内含有事务。若update失败，则回滚。
     * <p>
     * 不抛出异常。
     *
     * @param conn 数据源连接。
     * @param sql  update sql语句。
     * @return 是否执行成功。true：成功。false：失败。
     * @throws NodeException 如果执行update sql失败。
     */
    public static boolean executeUpdateSql(Connection conn, String sql, Object[] params) throws NodeException {
        QueryRunner queryRunner = new QueryRunner();
        try {
            conn.setAutoCommit(false);
            queryRunner.update(conn, sql, params);
            // d2a外部事务实现，如果开启，将数据库连接自动提交关闭
            if (!Constants.OPEN_GLOBAL_TRANSACTION.get()) {
                conn.commit();
            }
        } catch (SQLException e) {
            try {
                conn.rollback();
            } catch (SQLException e1) {
                log.error("[DSUtils] executeUpdateSql rollBack error.", e1);
                throw new NodeException("[DSUtils] executeUpdateSql rollBack error.", e);
            }
            log.error("[DSUtils] executeUpdateSql error. update sql:{}", sql, e);
            throw new NodeException("[DSUtils] executeUpdateSql error.", e);
        }
        return true;
    }

    public static <T> List<T> queryList(Connection conn, String sql, Object[] params,Class clazz) throws SQLException {
        QueryRunner queryRunner = new QueryRunner();
        List<T> list = null;
        try {
            list = (List)queryRunner.query(conn, sql, new BeanListHandler(clazz),params);
        } catch (SQLException e) {
            throw new SQLException("[DSUtils] queryList error. sql:{}", sql, e);
        }
        return list;
    }
}
