package net.wanji.common.tool.drools;

import com.alipay.jarslink.api.ModuleManager;
import net.wanji.common.framework.rest.JsonViewObject;
import net.wanji.common.framework.spring.ServiceBeanContext;
import net.wanji.common.utils.tool.StringUtils;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.drools.compiler.kie.builder.impl.InternalKieModule;
import org.drools.core.base.RuleNameStartsWithAgendaFilter;
import org.kie.api.KieServices;
import org.kie.api.builder.KieRepository;
import org.kie.api.builder.ReleaseId;
import org.kie.api.runtime.KieContainer;
import org.kie.api.runtime.KieSession;
import org.kie.api.runtime.rule.AgendaFilter;

import java.util.*;

/**
 * <>
 *
 * @author wanji
 * 2020/6/20 22:39
 */
@Slf4j
public class DroolsTool {

    /**
     * 标识符
     */
    private static final String GROUP_NAME = "com.unis.drools";
    /**
     * 版本
     */
    private static final String VERSION = "1.0.0";

    /**
     * 存放session  避免重复编译KieContainer
     */
    private static Map<String, KieContainer> kieContainers;

    /**
     * kie入口
     */
    private static KieServices kieServices;
    /**
     * KieModule仓库
     */
    private static KieRepository repository;

    static {
        kieServices = KieServices.Factory.get();
        repository = kieServices.getRepository();
        kieContainers = Maps.newConcurrentMap();
        System.setProperty("drools.dateformat", "yyyy-MM-dd");
    }

    /**
     * 初始化
     *
     * @param ruleCategoryId 规则目录id
     * @param scriptMap      规则脚本Map{ruleScriptId:script}
     */
    public static void init(String ruleCategoryId, Map<String, String> scriptMap) {
        if (StringUtils.isBlank(ruleCategoryId) || MapUtils.isEmpty(scriptMap)) {
            log.info("初始化规则参数不合法【ruleCategoryId:{}】", ruleCategoryId);
            return;
        }
        scriptMap.forEach((k, v) -> {
            loadRule(ruleCategoryId, k, v);
        });
    }

    /**
     * 初始化
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @param script         规则脚本
     */
    public static void init(String ruleCategoryId, String ruleScriptId, String script) {
        if (StringUtils.isBlank(ruleCategoryId) || StringUtils.isBlank(ruleScriptId) || StringUtils.isBlank(script)) {
            log.info("初始化规则参数为空");
            return;
        }
        loadRule(ruleCategoryId, ruleScriptId, script);
    }

    /**
     * 销毁session
     *
     * @param scriptMap 规则脚本Map{ruleScriptId:script}
     */
    public static void destroy(String ruleCategoryId, Map<String, String> scriptMap) {
        scriptMap.forEach((k, v) -> {
            if (StringUtils.isBlank(ruleCategoryId) || StringUtils.isBlank(k)) {
                log.info("销毁参数为空");
                return;
            }
            kieContainers.remove(genKey(ruleCategoryId, k));
        });
    }

    /**
     * 销毁session
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     */
    public static void destroy(String ruleCategoryId, String ruleScriptId) {
        if (StringUtils.isBlank(ruleCategoryId) || StringUtils.isBlank(ruleScriptId)) {
            log.info("销毁参数为空");
            return;
        }
        kieContainers.remove(genKey(ruleCategoryId, ruleScriptId));
    }

    /**
     * 执行单条
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @param fact           fact
     * @return List
     */
/*    public static List<Map<String, Object>> batchExecute(String ruleCategoryId, String ruleScriptId, Map<String, Object> fact) {
        return fireRules(ruleCategoryId, ruleScriptId, null, genRuleNameStartsWithAgendaFilter(ruleCategoryId, ruleScriptId), Collections.singletonList(fact));
    }*/
    public static JsonViewObject batchExecute(String ruleCategoryId, String ruleScriptId, Map<String, Object> fact) {
        return null;
    }
    /**
     * 执行多条（返回值过滤）
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @param facts          facts
     * @return List
     */
    /*public static List<Map<String, Object>> batchExecute(String ruleCategoryId, String ruleScriptId, List<Map<String, Object>> facts) {
        return fireRules(ruleCategoryId, ruleScriptId, null, genRuleNameStartsWithAgendaFilter(ruleCategoryId, ruleScriptId), facts);
    }*/

    public static JsonViewObject batchExecute(String ruleCategoryId, String ruleScriptId, List<Map<String, Object>> facts) {
        return null;
    }
    /**
     * 按规则流-执行多条（返回值过滤）
     *
     * @param ruleCategoryId   规则目录id
     * @param ruleScriptIdList 规则脚本id
     * @param facts            facts
     * @return List
     */
    public static List<Map<String, Object>> batchExecute(String ruleCategoryId, List<String> ruleScriptIdList, List<Map<String, Object>> facts) {
        List<Map<String, Object>> list = Lists.newArrayList();
        list.addAll(facts);
        for (String ruleScriptId : ruleScriptIdList) {
            List<Map<String, Object>> maps = fireRules(ruleCategoryId, ruleScriptId, null, genRuleNameStartsWithAgendaFilter(ruleCategoryId, ruleScriptId), list);
            list.addAll(maps);
        }
        return list;
    }

    /**
     * 生成议程组
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @return String
     */
    private static String genAgendaGroup(String ruleCategoryId, String ruleScriptId) {
        return String.format("agendaGroup%s%s", ruleCategoryId, ruleScriptId);
    }

    /**
     * 获取规则前缀过滤
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @return AgendaFilter 议程组过滤
     */
    public static AgendaFilter genRuleNameStartsWithAgendaFilter(String ruleCategoryId, String ruleScriptId) {
        return new RuleNameStartsWithAgendaFilter(DroolsTool.genRuleName(ruleCategoryId, ruleScriptId));
    }

    /**
     * 规则名
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @return String
     */
    public static String genRuleName(String ruleCategoryId, String ruleScriptId) {
        return genRuleName(ruleCategoryId, ruleScriptId, "");
    }

    /**
     * 规则名
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @param suffix         后缀
     * @return String
     */
    public static String genRuleName(String ruleCategoryId, String ruleScriptId, String suffix) {
        return String.format("rule%s%s%s", ruleCategoryId, ruleScriptId, suffix);
    }

    /**
     * 执行指定数据集下的所有AgendaGroup指定议程组和“MAIN”议程组(使用agenda-group定义，默认是MAIN)的并且
     * 经过AgendaFilter过滤的规则
     *
     * @param ruleCategoryId 规则所在的目录
     * @param agendaGroup    规则所在议程组名称，如果不定义默认：MAIN
     * @param filter         规则过滤器
     * @param facts          事实
     */
    private static List<Map<String, Object>> fireRules(String ruleCategoryId, String ruleScriptId, String agendaGroup, AgendaFilter filter, List<Map<String, Object>> facts) {
        KieSession session = null;
        try {
            session = newKieSession(ruleCategoryId, ruleScriptId);
            session.setGlobal("moduleManager", ServiceBeanContext.getBean(ModuleManager.class));
            // 新增 fact
            for (Object fact : facts) {
                session.insert(fact);
            }
            Set<Map<String, Object>> set = Sets.newConcurrentHashSet();
            session.insert(set);
            //focus agenda group
            if (agendaGroup != null && !agendaGroup.isEmpty()) {
                //执行分组
                session.getAgenda().getAgendaGroup(agendaGroup).setFocus();
            }
            // 执行
            int count;
            if (filter != null) {
                count = session.fireAllRules(filter);
            } else {
                count = session.fireAllRules();
            }
            log.info("执行并匹配到的规则数量:{}", count);
            return new ArrayList<>(new HashSet<>(set));
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage());
        } finally {
            if (session != null) {
                session.dispose();
            }
        }
    }

    /**
     * 加载规则
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     */
    private static void loadRule(String ruleCategoryId, String ruleScriptId, String script) {
        String artifactId = String.format("unis-%s", ruleCategoryId);
        ReleaseId releaseId = kieServices.newReleaseId(GROUP_NAME, artifactId, VERSION);
        InternalKieModule kieModule = DroolsUtils.createKieModule(kieServices, releaseId, ruleCategoryId, ruleScriptId, script);
        repository.addKieModule(kieModule);
        KieContainer kieContainer = kieServices.newKieContainer(releaseId);
        kieContainer.updateToVersion(releaseId);
        kieContainers.put(genKey(ruleCategoryId, ruleScriptId), kieContainer);
    }

    /**
     * 创建指定数据集下的KieSession
     *
     * @param ruleCategoryId 规则目录主键
     * @return KieSession
     */
    private static KieSession newKieSession(String ruleCategoryId, String ruleScriptId) {
        String key = genKey(ruleCategoryId, ruleScriptId);
        KieContainer kieContainer = kieContainers.get(key);
        if (kieContainer == null) {
            throw new RuntimeException("无法获取KieContainer:" + key);
        }
        return kieContainer.newKieSession();
    }

    /**
     * 获取容器Map的key
     *
     * @param ruleCategoryId 规则目录id
     * @return String
     */
    private static String genKey(String ruleCategoryId, String ruleScriptId) {
        return String.format("kie-rule-category-%s-%s", ruleCategoryId, ruleScriptId);
    }

    /**
     * 判断key是否存在
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @return boolean
     */
    public static boolean exists(String ruleCategoryId, String ruleScriptId) {
        return kieContainers.containsKey(genKey(ruleCategoryId, ruleScriptId));
    }

}
