package net.wanji.common.tool.drools;

import net.wanji.common.utils.tool.StringUtils;
import org.drools.compiler.kie.builder.impl.InternalKieModule;
import org.kie.api.KieServices;
import org.kie.api.builder.KieBuilder;
import org.kie.api.builder.KieFileSystem;
import org.kie.api.builder.ReleaseId;
import org.kie.api.builder.model.KieBaseModel;
import org.kie.api.builder.model.KieModuleModel;
import org.kie.api.builder.model.KieSessionModel;
import org.kie.api.conf.EqualityBehaviorOption;
import org.kie.api.conf.EventProcessingOption;

/**
 * <>
 *
 * @author wanji
 * 2020/6/4 13:42
 */
public class DroolsUtils {

    private static final String ROOT = "src/main/resources/";

    /**
     * 创建kieModule
     *
     * @param kieServices
     * @param releaseId
     * @param ruleScriptId
     * @param script
     * @return
     */
    public static InternalKieModule createKieModule(KieServices kieServices, ReleaseId releaseId, String ruleCategoryId, String ruleScriptId, String script) {
        KieFileSystem kfs = createKieFileSystemWithKProject(kieServices, true);
        kfs.writePomXML(genPom(releaseId));
        if (StringUtils.isNotBlank(script)) {
            kfs.write(genDrlPath(ruleCategoryId, ruleScriptId), script);
        }
        KieBuilder kieBuilder = kieServices.newKieBuilder(kfs);
        kieBuilder.buildAll();
        if (!kieBuilder.getResults().getMessages().isEmpty()) {
            throw new IllegalStateException(
                    "Error creating KieBuilder. errorMsg:" + kieBuilder.getResults().getMessages());
        }
        return (InternalKieModule) kieBuilder.getKieModule();
    }

    /**
     * 生成drl虚拟路径
     *
     * @param ruleCategoryId 规则目录id
     * @param ruleScriptId   规则脚本id
     * @return String
     */
    private static String genDrlPath(String ruleCategoryId, String ruleScriptId) {
        return String.format("%s%s/%s.drl", ROOT, ruleCategoryId, ruleScriptId);
    }

    /**
     * 创建默认的kbase和stateful的kiesession
     *
     * @param kieServices
     * @param isdefault   是否默认
     * @return KieFileSystem
     */
    private static KieFileSystem createKieFileSystemWithKProject(KieServices kieServices, boolean isdefault) {
        //对应kmodule节点
        KieModuleModel kieModuleModel = kieServices.newKieModuleModel();
        //对应kbase节点
        KieBaseModel kieBaseModel = kieModuleModel
                .newKieBaseModel("KBase")
                .setDefault(isdefault)
                .setEqualsBehavior(EqualityBehaviorOption.EQUALITY)
                .setEventProcessingMode(EventProcessingOption.STREAM);
        // Configure the KieSession.
        kieBaseModel.newKieSessionModel("KSession")
                .setDefault(isdefault)
                .setType(KieSessionModel.KieSessionType.STATEFUL);
        KieFileSystem kfs = kieServices.newKieFileSystem();
        kfs.writeKModuleXML(kieModuleModel.toXML());
        return kfs;
    }

    /**
     * 生成pom.xml
     *
     * @param releaseId    发布id
     * @param dependencies 依赖
     * @return String
     */
    private static String genPom(ReleaseId releaseId, ReleaseId... dependencies) {
        String pom = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                + "<project xmlns=\"http://maven.apache.org/POM/4.0.0\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
                + "         xsi:schemaLocation=\"http://maven.apache.org/POM/4.0.0 http://maven.apache.org/maven-v4_0_0.xsd\">\n"
                + "  <modelVersion>4.0.0</modelVersion>\n" + "\n" + "  <groupId>" + releaseId.getGroupId()
                + "</groupId>\n" + "  <artifactId>" + releaseId.getArtifactId() + "</artifactId>\n" + "  <version>"
                + releaseId.getVersion() + "</version>\n" + "\n";
        if (dependencies != null && dependencies.length > 0) {
            pom += "<dependencies>\n";
            for (ReleaseId dep : dependencies) {
                pom += "<dependency>\n";
                pom += "  <groupId>" + dep.getGroupId() + "</groupId>\n";
                pom += "  <artifactId>" + dep.getArtifactId() + "</artifactId>\n";
                pom += "  <version>" + dep.getVersion() + "</version>\n";
                pom += "</dependency>\n";
            }
            pom += "</dependencies>\n";
        }
        pom += "</project>";
        return pom;
    }


}
