package net.wanji.common.tool.estool;

import com.alibaba.fastjson.JSON;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.map.HashedMap;
import org.elasticsearch.action.bulk.BulkRequestBuilder;
import org.elasticsearch.action.bulk.BulkResponse;
import org.elasticsearch.action.delete.DeleteResponse;
import org.elasticsearch.action.get.GetResponse;
import org.elasticsearch.action.get.MultiGetItemResponse;
import org.elasticsearch.action.get.MultiGetResponse;
import org.elasticsearch.action.search.SearchRequestBuilder;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.update.UpdateRequest;
import org.elasticsearch.client.transport.TransportClient;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.common.transport.TransportAddress;
import org.elasticsearch.common.xcontent.XContentBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.sort.SortBuilder;
import org.elasticsearch.transport.client.PreBuiltTransportClient;

import java.net.InetAddress;
import java.util.*;

import static org.elasticsearch.common.xcontent.XContentFactory.jsonBuilder;

/**
 * 对ES操作的工具类
 *
 * @author wanji
 * @date 2016/11/23
 */
@Slf4j
public class EsUtil {
    private static EsUtil esUtil = null;

    public static EsUtil getInstance() {
        if (esUtil == null) {
            esUtil = new EsUtil();
        }
        return esUtil;
    }

    /**
     * 测试连接是否存在
     *
     * @param clusterName
     * @param address
     * @return
     */
    public boolean checkConn(String clusterName, Map<String, Integer> address) {
        boolean flag = false;
        TransportClient client = null;
        try {
            Settings settings = Settings.builder()
                    .put("cluster.name", clusterName)
                    .put("client.transport.sniff", false)
                    .build();
            client = new PreBuiltTransportClient(settings);
            for (Map.Entry<String, Integer> map : address.entrySet()) {
                client.addTransportAddress(new TransportAddress(InetAddress.getByName(map.getKey()), map.getValue()));
            }
            if (client.connectedNodes().size() > 0) {
                flag = true;
            }
        } catch (Exception e) {
            log.error("ES-Server Connection is error!", e);
        } finally {
            if (null != client) {
                client.close();
            }
        }
        return flag;
    }

    /**
     * 得到ES连接
     *
     * @return
     */
    public TransportClient esClient(String clusterName, Map<String, Integer> address) {
        TransportClient client = null;
        try {
            Settings settings = Settings.builder()
                    .put("cluster.name", clusterName)
                    .put("client.transport.sniff", false)
                    .build();
            client = new PreBuiltTransportClient(settings);
            for (Map.Entry<String, Integer> map : address.entrySet()) {
                client.addTransportAddress(new TransportAddress(InetAddress.getByName(map.getKey()), map.getValue()));
            }
        } catch (Exception e) {
            log.error("ES-Server Connection is error!", e);
        }
        return client;
    }


    /**
     * 关闭连接
     *
     * @param client
     */
    public void closeConn(TransportClient client) {
        if (null != client) {
            client.close();
        }
    }

    /**
     * 创建索引
     *
     * @param client
     * @param indexName
     * @param indexNumberShards
     * @param indexNumberReplicas
     */
    public boolean createIndex(TransportClient client, String indexName, Long indexNumberShards, Long indexNumberReplicas) {
        boolean flag = true;
        try {
            client.admin().indices().prepareCreate(indexName)
                    .setSettings(Settings.builder()
                            .put("index.number_of_shards", indexNumberShards)
                            .put("index.number_of_replicas", indexNumberReplicas)
                    )
                    .get();
        } catch (Exception e) {
            flag = false;
            log.error("[indexName:" + indexName + "],[indexNumberShards:" + indexNumberShards + "],[indexNumberReplicas:" + indexNumberReplicas + "]  EsUtil  createIndex is error! ", e);
        }
        return flag;
    }

    /**
     * 删除索引
     *
     * @param client
     * @param indexName
     * @return
     */
    public boolean deleteIndex(TransportClient client, String indexName) {
        boolean flag = true;
        try {
            client.admin().indices().prepareDelete(indexName).execute().actionGet();
        } catch (Exception e) {
            flag = false;
            log.error("[indexName:" + indexName + "]  EsUtil  createIndex is error! ", e);
        }
        return flag;
    }

    /**
     * 创建Mapping
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param map
     * @return
     */
    public boolean createMapping(TransportClient client, String indexName, String typeName, Map<String, String> map) {
        boolean flag = true;
        Map<String, Map<String, Map<String, String>>> properMap = Maps.newHashMap();
        try {
            Map<String, Map<String, String>> propMap = Maps.newHashMap();
            for (Map.Entry<String, String> entry : map.entrySet()) {
                Map<String, String> typeMap = Maps.newHashMap();
                typeMap.put("type", entry.getValue());
                propMap.put(entry.getKey(), typeMap);
            }
            properMap.put("properties", propMap);
            String json = JSON.toJSONString(properMap);
            client.admin().indices().preparePutMapping(indexName)
                    .setType(typeName)
                    .setSource(json)
                    .get();
        } catch (Exception e) {
            flag = false;
            log.error("EsUtil  createMapping is error!", e);
        }
        return flag;
    }

    /**
     * 删除mapping
     *
     * @param client
     * @param index
     * @param typeName
     * @return
     * @throws Exception
     */
    public boolean deleteMapping(TransportClient client, String index, String typeName) throws Exception {


        return true;
    }


    /**
     * 向指定的索引库中添加数据
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param map
     * @return
     */
    public boolean putIndexData(TransportClient client, String indexName, String typeName, String id, Map<String, Object> map) {
        boolean flag = true;
        XContentBuilder builder;
        try {
            builder = jsonBuilder().startObject();
            for (Map.Entry<String, Object> mapData : map.entrySet()) {
                builder.field(mapData.getKey(), mapData.getValue());
            }
            builder.endObject();
            client.prepareIndex(indexName, typeName, id)
                    .setSource(builder)
                    .get();
            log.info("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "] put succeed!");
        } catch (Exception e) {
            flag = false;
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "]  EsUtil  findDataById is error! ", e);
        }
        return flag;
    }

    /**
     * 批量添加数据
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param datas
     * @return
     */
    public boolean putIndexDatas(TransportClient client, String indexName, String typeName, List<Map<String, Object>> datas) {
        boolean flag = true;
        BulkResponse bulkResponse = null;
        try {
            BulkRequestBuilder bulkRequest = client.prepareBulk();
            for (Map<String, Object> data : datas) {
                data.remove("id");
                bulkRequest.add(client.prepareIndex(indexName, typeName, (String) data.get("id"))
                        .setSource(data)
                );
            }

            bulkResponse = bulkRequest.get();
            if (bulkResponse.hasFailures()) {
                log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[Item:" + bulkResponse.getItems() + "] put fail!");
            }
            log.debug("[indexName:" + indexName + "],[typeName:" + typeName + "],[Item:" + bulkResponse.getItems() + "] put succeed!");
        } catch (Exception e) {
            flag = false;
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[Item:" + bulkResponse.getItems() + "]  EsUtil  putIndexDatas is error! ", e);
        }
        return flag;
    }

    /**
     * 根据索引的Id查找索引的数据
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param id
     */
    public Map<String, Object> findDataById(TransportClient client, String indexName, String typeName, String id) {
        Map<String, Object> map = new HashMap<>();
        try {
            GetResponse res = client.prepareGet(indexName, typeName, id).get();
            map = res.getSourceAsMap();
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "] EsUtil  findDataById is error! ", e);
        }
        return map;
    }

    /**
     * 根据多个Id查询多条数据
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param ids
     * @return
     */
    public String findDataByIds(TransportClient client, String indexName, String typeName, String... ids) {
        String json = "";
        try {
            MultiGetResponse multiGetItemResponses = client.prepareMultiGet()
                    .add(indexName, typeName, ids)
                    .get();
            for (MultiGetItemResponse itemResponse : multiGetItemResponses) {
                GetResponse response = itemResponse.getResponse();
                if (response.isExists()) {
                    json = response.getSourceAsString();
                }
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[ids:" + ids + "] EsUtil  findDataByIds is error! ", e);
        }
        return json;
    }

    /**
     * 根据关键字进行查询
     *
     * @param client
     * @return
     */
    public List<Map<String, Object>> searchDatas(TransportClient client, String indexName, String typeName, String key, String value) {
        List<Map<String, Object>> list = Lists.newArrayList();
        try {
            SearchResponse response = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(QueryBuilders.matchQuery(key, value))
                    .execute()
                    .actionGet();
            for (SearchHit hit : response.getHits().getHits()) {
                hit.getSourceAsMap().put("id", hit.getId());
                list.add(hit.getSourceAsMap());
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "],[value:" + value + "] EsUtil  searchDatas is error! ", e);
        }
        return list;
    }

    /**
     * 组合查询
     * QueryBuilders
     * .boolQuery()
     * .must(QueryBuilders.matchQuery("name", "value"))
     * .must(QueryBuilders.matchQuery("home", "value"))
     * .mustNot(QueryBuilders.matchQuery("isRealMen", false))
     * .should(QueryBuilders.matchQuery("now_home", "value"));
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param qb
     * @return
     */
    public List<Map<String, Object>> searchDatas(TransportClient client, String indexName, String typeName, QueryBuilder qb) {
        List<Map<String, Object>> list = Lists.newArrayList();
        try {
            SearchResponse response = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(qb)
                    .setExplain(true)
                    .execute()
                    .actionGet();
            for (SearchHit hit : response.getHits().getHits()) {
                hit.getSourceAsMap().put("id", hit.getId());
                list.add(hit.getSourceAsMap());
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "], EsUtil  searchDatas is error! ", e);
        }
        return list;
    }

    /**
     * 分页查询
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param qb        可以为null     表示没有条件
     * @param startRow
     * @param pageSize
     * @return
     */
    public List<Map<String, Object>> searchDatas(TransportClient client, String indexName, String typeName, QueryBuilder qb, int startRow, int pageSize) {
        List<Map<String, Object>> list = Lists.newArrayList();
        try {
            SearchResponse response = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(qb)
                    .setFrom(startRow).setSize(pageSize)
                    .setExplain(true)
                    .execute()
                    .actionGet();
            for (SearchHit hit : response.getHits().getHits()) {
                hit.getSourceAsMap().put("id", hit.getId());
                list.add(hit.getSourceAsMap());
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "], EsUtil  searchDatas is error! ", e);
        }
        return list;
    }

    public List<Map<String, Object>> searchDatas(TransportClient client, String indexName, String typeName, QueryBuilder qb, Queue<SortBuilder> sortBuilders) {
        List<Map<String, Object>> list = Lists.newArrayList();
        try {
            SearchRequestBuilder request = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(qb)
                    .setExplain(true);
            for (SortBuilder sort : sortBuilders) {
                request.addSort(sort);
            }
            SearchResponse response = request.execute().actionGet();
            for (SearchHit hit : response.getHits().getHits()) {
                hit.getSourceAsMap().put("id", hit.getId());
                list.add(hit.getSourceAsMap());
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "], EsUtil  searchDatas is error! ", e);
        }
        return list;
    }

    public List<Map<String, Object>> searchDatas(TransportClient client, String indexName, String typeName, QueryBuilder qb, int startRow, int pageSize, Queue<SortBuilder> sortBuilders) {
        List<Map<String, Object>> list = Lists.newArrayList();
        try {
            SearchRequestBuilder request = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(qb)
                    .setFrom(startRow).setSize(pageSize)
                    .setExplain(true);
            for (SortBuilder sort : sortBuilders) {
                request.addSort(sort);
            }
            SearchResponse response = request.execute().actionGet();
            for (SearchHit hit : response.getHits().getHits()) {
                hit.getSourceAsMap().put("id", hit.getId());
                list.add(hit.getSourceAsMap());
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "], EsUtil  searchDatas is error! ", e);
        }
        return list;
    }


    /**
     * @param client
     * @param indexName 索引名称
     * @param typeName  类型名称
     * @param qb        查询条件
     * @param startRow  分页起始行
     * @param pageSize  分页终止行
     * @param fileds    需要返回的字段
     * @return java.util.List<java.util.Map                               <                               java.lang.String                               ,                               java.lang.Object>>
     * @Description: 分页查询，并返回指定字段
     * @Author wanji
     * @Date 2019/12/16 16:34
     **/
    public List<Map<String, Object>> searchDatas(TransportClient client, String indexName, String typeName, QueryBuilder qb, int startRow, int pageSize, Set<String> fileds) {
        List<Map<String, Object>> list = Lists.newArrayList();
        if (null == fileds || fileds.isEmpty()) {
            return list;
        }
        try {
            SearchResponse response = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(qb)
                    .setFrom(startRow).setSize(pageSize)
                    .setExplain(true)
                    .execute()
                    .actionGet();
            for (SearchHit hit : response.getHits().getHits()) {
                Map<String, Object> result = new HashMap<>(response.getHits().getHits().length);
                fileds.retainAll(hit.getSourceAsMap().keySet());
                if (fileds.isEmpty()) {
                    return list;
                }
                result.put("id", hit.getId());
                for (String field : fileds) {
                    result.put(field, hit.getSourceAsMap().get(field));
                }
                list.add(result);
            }
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "], EsUtil  searchDatas is error! ", e);
        }
        return list;
    }

    /**
     * 获取符合条件的数据总数
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param qb
     * @return
     */
    public Long searchDatasCount(TransportClient client, String indexName, String typeName, QueryBuilder qb) {
        long count = 0L;
        try {
            SearchResponse response = client.prepareSearch(indexName).setTypes(typeName)
                    .setQuery(qb)
                    .setExplain(true)
                    .execute()
                    .actionGet();
            count = response.getHits().getTotalHits();
        } catch (Exception e) {
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[name:" + typeName + "], EsUtil  searchDatasCount is error! ", e);
        }
        return count;
    }


    /**
     * 根据索引的Id删除数据
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param id
     * @return
     */
    public boolean deleteDataById(TransportClient client, String indexName, String typeName, String id) {
        boolean flag = true;
        try {
            DeleteResponse response = client.prepareDelete(indexName, typeName, id).get();
            log.info("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "],[status:" + response.status() + "] delete succeed!");
        } catch (Exception e) {
            flag = false;
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "] EsUtil  deleteDataById is error! ", e);
        }
        return flag;
    }

    /**
     * 根据Id修改索引的数据
     *
     * @param client
     * @param indexName
     * @param typeName
     * @param id
     * @param map
     * @return
     */
    public boolean updateDataById(TransportClient client, String indexName, String typeName, String id, Map<String, Object> map) {
        boolean flag = true;
        try {
            XContentBuilder builder = jsonBuilder().startObject();
            for (Map.Entry<String, Object> mapData : map.entrySet()) {
                builder.field(mapData.getKey(), mapData.getValue());
            }
            builder.endObject();

            UpdateRequest updateRequest = new UpdateRequest(indexName, typeName, id).doc(builder);
            client.update(updateRequest).get();
            log.info("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "],[map:" + map + "] update succeed!");
        } catch (Exception e) {
            flag = false;
            log.error("[indexName:" + indexName + "],[typeName:" + typeName + "],[id:" + id + "],[map:" + map + "]  EsUtil  updateDataById is error! ", e);
        }
        return flag;
    }


    public static void main(String[] arg) {
        EsUtil esUtil = new EsUtil();
        try {
            Map<String, Integer> map = new HashedMap();
            map.put("219.142.87.89", 9300);
//            map.put("192.168.5.10",9300);
            TransportClient client = EsUtil.getInstance().esClient("business-platform", map);

            Map<String, Object> dataMap = Maps.newHashMap();
            dataMap.put("filed1", "name");
            dataMap.put("filed2", 10);
            dataMap.put("filed3", 1);
            dataMap.put("filed4", 1.1);
            dataMap.put("filed5", 1.001);
            dataMap.put("filed6", true);
            dataMap.put("filed7", new Date());

            Map<String, String> mapping = Maps.newHashMap();
            mapping.put("filed1", "string");
            mapping.put("filed2", "integer");
            mapping.put("filed3", "long");
            mapping.put("filed4", "float");
            mapping.put("filed5", "double");
            mapping.put("filed6", "boolean");
            mapping.put("filed7", "date");

            esUtil.deleteIndex(client, "dev_service_plugin_log");

            esUtil.deleteMapping(client, "dev_service_plugin_log", "dev_service_plugin_log");
            //       esUtil.searchDatas(client,"vim","employee","name","hudw");
            //         esUtil.createMapping(client,"vim","mapping",mapping);

//
            //esUtil.putIndexData(client,"test2","delet_test1","2",data_map);
//             System.out.println(esUtil.findDataById(client,"test2","delet_test1","2"));
//            esUtil.updateDataById(client,"test_index","test1","2",data_map);
//            System.out.println(esUtil.findDataById(client,"test_index","test1","2"));
        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
