package net.wanji.common.tool.mqtool;

import com.alibaba.fastjson.JSONObject;
import net.wanji.common.framework.Constants;
import net.wanji.common.framework.spring.ServiceBeanContext;
import net.wanji.common.framework.thread.NamedThreadFactory;
import net.wanji.common.utils.tool.StringUtils;
import com.rabbitmq.client.*;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;

/**
 * RabbitMQ 收发消息统一帮助类
 *
 * @author gl.dong
 * @date 2020/7/1
 */
@Slf4j
@Data
public class MqUtil {

    /**
     * 消息发送线程池
     */
    private static final ExecutorService EXECUTOR = Executors.newSingleThreadExecutor(
            new NamedThreadFactory("gisc-message-thread", true));

    /**
     * RabbitMq连接工厂对象
     **/
    private volatile static ConnectionFactory factory = null;

    /**
     * 默认的队列名
     */
    private final static String DEFAULT_QUEUE_NAME = "GISC-TOPIC-LOG";

    public MqUtil() {
    }

    /**
     * 获取RabbitMQ的连接
     *
     * @return
     */
    public static ConnectionFactory getInstance() throws Exception {
        if (factory == null) {
            /**
             * 服务器域名或IP
             */
            String host = ServiceBeanContext.getContext().getEnvironment().getProperty("rabbitmq.host");
            /**
             * 端口，默认5672
             */
            String port = ServiceBeanContext.getContext().getEnvironment().getProperty("rabbitmq.port");
            /**
             * 用户名，默认guest
             */
            String username = ServiceBeanContext.getContext().getEnvironment().getProperty("rabbitmq.username");
            /**
             * 用户密码，默认guest
             */
            String password = ServiceBeanContext.getContext().getEnvironment().getProperty("rabbitmq.password");
            /**
             * 虚拟目录，默认/
             */
            String virtualHost = ConnectionFactory.DEFAULT_VHOST;

            if (!StringUtils.isNotBlank(host, username, password, virtualHost)) {
                log.error("MQ初始配置参数错误[host:{},virtualHost:{}]", host, virtualHost);
                throw new Exception("rabbitMQ初始化参数异常");
            }
            if (StringUtils.isBlank(port)) {
                port = ConnectionFactory.DEFAULT_AMQP_PORT + "";
                log.info("MQ初始配置参数启用默认端口");
            }

            factory = new ConnectionFactory();
            //设置服务地址
            factory.setHost(host);
            //端口
            factory.setPort(Integer.parseInt(port));
            //设置账号信息，用户名、密码、vhost
            //test 设置虚拟机，一个mq服务可以设置多个虚拟机，每个虚拟机就相当于一个独立的mq
            factory.setVirtualHost(virtualHost);
            factory.setUsername(username);
            factory.setPassword(password);
        }
        return factory;
    }

    /**
     * 创建交换机
     * <p>
     * exchange  交换器名称
     * type  交换器类型 DIRECT("direct"), FANOUT("fanout"), TOPIC("topic"), HEADERS("headers");
     * durable  是否持久化,durable设置为true表示持久化,反之是非持久化,持久化的可以将交换器存盘,在服务器重启的时候不会丢失信息.
     * autoDelete  是否自动删除,设置为TRUE则表是自动删除,自删除的前提是至少有一个队列或者交换器与这交换器绑定,之后所有与这个交换器绑定的队列或者交换器都与此解绑,一般都设置为false
     * internal  是否内置,如果设置 为true,则表示是内置的交换器,客户端程序无法直接发送消息到这个交换器中,只能通过交换器路由到交换器的方式
     * arguments  其它一些结构化参数比如:alternate-exchange
     */
    private static void exchangeDeclare(String exchangeName, String type, boolean durable, boolean autoDelete, boolean internal, Map<String, Object> arguments) {
        Connection connection = null;
        Channel channel = null;
        try {
            if (StringUtils.isNotBlank(exchangeName)) {
                // 1、通过工厂获取连接
                connection = MqUtil.getInstance().newConnection();
                // 2、从连接中创建通道，使用通道才能完成消息相关的操作
                channel = connection.createChannel();
                // 3、 创建 exchange
                channel.exchangeDeclare(exchangeName, type, durable, autoDelete, internal, arguments);
            }
        } catch (Exception e) {
            log.error("交换机创建失败 exchangeName:{} , error:{}", exchangeName, e.getMessage());
        } finally {
            close(connection, channel);
        }
    }

    /**
     * internal : false
     */
    private static void exchangeDeclare(String exchangeName, String type, boolean durable, boolean autoDelete, Map<String, Object> arguments) {
        exchangeDeclare(exchangeName, type, durable, autoDelete, false, arguments);
    }

    /**
     * durable : false
     * autoDelete : false
     * internal : false
     * arguments : null
     */
    private static void exchangeDeclare(String exchangeName, String type) {
        exchangeDeclare(exchangeName, type, false, false, null);
    }

    /**
     * type : topic
     * durable : false
     * autoDelete : false
     * internal : false
     * arguments : null
     */
    private static void exchangeDeclare(String exchangeName) {
        exchangeDeclare(exchangeName, "topic");
    }

    /**
     * 异步发送消息
     *
     * @param exchangeName 交换机名
     * @param topicName    消息队列名key
     * @param content      消息体
     */
    public static void sendMessage(String exchangeName, String topicName, String content) {
        EXECUTOR.execute(() -> {
            Connection connection = null;
            Channel channel = null;
            try {
                // 1、通过工厂获取连接
                connection = MqUtil.getInstance().newConnection();
                // 2、从连接中创建通道，使用通道才能完成消息相关的操作
                channel = connection.createChannel();
                // 3、 创建 exchange
                exchangeDeclare(exchangeName);
                commonSendMessage(channel, exchangeName, topicName, content);
            } catch (Exception e) {
                log.error("{} sendMessage is error", MqUtil.class.getSimpleName(), e);
            } finally {
                close(connection, channel);
            }
        });
    }

    public static boolean sendTopicMessage(String exchangeName, String topicName, String content) {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1、通过工厂获取连接
            connection = MqUtil.getInstance().newConnection();
            // 2、从连接中创建通道，使用通道才能完成消息相关的操作
            channel = connection.createChannel();
            // 3、 创建 exchange
            exchangeDeclare(exchangeName);
            commonSendMessage(channel, exchangeName, topicName, content);
            return true;
        } catch (Exception e) {
            log.error("{} sendLogMessage is error", MqUtil.class.getSimpleName(), e);
            return false;
        } finally {
            close(connection, channel);
        }
    }


    public static void receiveTopicMessage(String queueName,String keyName, MessageHandler handler, boolean isAutoAck) throws Exception {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1.获取到连接
            connection = MqUtil.getInstance().newConnection();
            // 2. 创建会话通道,生产者和mq服务所有通信都在channel通道中完成
            channel = connection.createChannel();
            /**
             * 3. 声明队列
             * 参数：String queue, boolean durable, boolean exclusive, boolean autoDelete, Map<String, Object> arguments
             * 参数明细
             * 1、queue 队列名称
             * 2、durable 是否持久化，如果持久化，mq重启后队列还在
             * 3、exclusive 是否独占连接，队列只允许在该连接中访问，如果connection连接关闭队列则自动删除,如果将此参数设置true可用于临时队列的创建
             * 4、autoDelete 自动删除，队列不再使用时是否自动删除此队列，如果将此参数和exclusive参数设置为true就可以实现临时队列（队列不用了就自动删除）
             * 5、arguments 参数，可以设置一个队列的扩展参数，比如：可设置存活时间
             */
            channel.queueDeclare(queueName, false, false, false, null);

            // 4. 实现消费方法
            commonReceive(channel, queueName, handler, isAutoAck);
        } catch (Exception e) {
            log.error("MQ 消费者启动失败 :{}", e);
        }
    }

    public static void receiveTopicMessage(String queueName, MessageHandler handler, boolean isAutoAck) throws Exception {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1.获取到连接
            connection = MqUtil.getInstance().newConnection();
            // 2. 创建会话通道,生产者和mq服务所有通信都在channel通道中完成
            channel = connection.createChannel();
            /**
             * 3. 声明队列
             * 参数：String queue, boolean durable, boolean exclusive, boolean autoDelete, Map<String, Object> arguments
             * 参数明细
             * 1、queue 队列名称
             * 2、durable 是否持久化，如果持久化，mq重启后队列还在
             * 3、exclusive 是否独占连接，队列只允许在该连接中访问，如果connection连接关闭队列则自动删除,如果将此参数设置true可用于临时队列的创建
             * 4、autoDelete 自动删除，队列不再使用时是否自动删除此队列，如果将此参数和exclusive参数设置为true就可以实现临时队列（队列不用了就自动删除）
             * 5、arguments 参数，可以设置一个队列的扩展参数，比如：可设置存活时间
             */
            channel.queueDeclare(queueName, false, false, false, null);

            // 4. 实现消费方法
            commonReceive(channel, queueName, handler, isAutoAck);
        } catch (Exception e) {
            log.error("MQ 消费者启动失败 :{}", e);
        }
    }
    /**
     * 同步发送消息
     *
     * @param exchangeName 交换机名
     * @param topicName    消息队列名key
     * @param content      消息体
     */
    public static boolean sendSyncMessage(String exchangeName, String topicName, String content) {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1、通过工厂获取连接
            connection = MqUtil.getInstance().newConnection();
            // 2、从连接中创建通道，使用通道才能完成消息相关的操作
            channel = connection.createChannel();
            // 3、 创建 exchange
            exchangeDeclare(exchangeName);
            commonSendMessage(channel, exchangeName, topicName, content);
            return true;
        } catch (Exception e) {
            log.error("{} sendLogMessage is error", MqUtil.class.getSimpleName(), e);
            return false;
        } finally {
            close(connection, channel);
        }
    }

    private static void commonSendMessage(Channel channel, String exchangeName, String topicName, String content) throws Exception {
        /**
         * 4、声明（创建）队列
         * 参数：String queue, boolean durable, boolean exclusive, boolean autoDelete, Map<String, Object> arguments
         * 参数明细
         * 1、queue 队列名称
         * 2、durable 是否持久化，如果持久化，mq重启后队列还在
         * 3、exclusive 是否独占连接，队列只允许在该连接中访问，如果connection连接关闭队列则自动删除,如果将此参数设置true可用于临时队列的创建
         * 4、autoDelete 自动删除，队列不再使用时是否自动删除此队列，如果将此参数和exclusive参数设置为true就可以实现临时队列（队列不用了就自动删除）
         * 5、arguments 参数，可以设置一个队列的扩展参数，比如：可设置存活时间
         */
        channel.queueDeclare(topicName, false, false, false, null);
        /**
         * 5、向指定的队列中发送消息
         *
         * 参数：String exchange, String routingKey, BasicProperties props, byte[] body
         * 参数明细：
         * 1、exchange，交换机，如果不指定将使用mq的默认交换机（设置为""）
         * 2、routingKey，路由key，交换机根据路由key来将消息转发到指定的队列，如果使用默认交换机，routingKey设置为队列的名称
         * 3、props，消息的属性
         * 4、body，消息内容
         */
        channel.basicPublish(exchangeName, topicName, null, content.getBytes());
        log.info(" MQ Send {}", content);
    }

    /**
     * 广播模式
     *
     * @param exchangeName 交换机名
     * @param topicName    消息队列名key
     * @param content      消息体
     */
    public static boolean sendFanoutMessage(String exchangeName,  String content) {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1、通过工厂获取连接
            connection = MqUtil.getInstance().newConnection();
            // 2、从连接中创建通道，使用通道才能完成消息相关的操作
            channel = connection.createChannel();
            // 3、 创建 exchange
            exchangeDeclare(exchangeName, "fanout");

            channel.basicPublish(exchangeName, exchangeName, null, content.getBytes());
            log.info(" MQ Send {}", content);
            return true;
        } catch (Exception e) {
            log.error("{} sendLogMessage is error", MqUtil.class.getSimpleName(), e);
            return false;
        } finally {
            close(connection, channel);
        }
    }


    public static void sendLogMessage(String indexName, String typeName, Map<String, Object> logData) {
        // 封装消息内容结
        Map<String, Object> map = new HashMap<>();
        map.put("type", Constants.BP_LOG);
        Map<String, Object> contentMap = new HashMap<>();
        contentMap.put("indexName", indexName);
        contentMap.put("typeName", typeName);
        contentMap.put("logData", logData);
        map.put("content", contentMap);
        String message = JSONObject.toJSONString(map);
        sendMessage("", Constants.BP_LOG, message);
    }

    /**
     * 接收处理消息
     *
     * @param queueName 消息队列名key
     * @param handler   消息处理类
     * @param isAutoAck 是否自动ACK
     * @throws Exception
     */
    public static void receiveMessage(String queueName, MessageHandler handler, boolean isAutoAck) throws Exception {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1.获取到连接
            connection = MqUtil.getInstance().newConnection();
            // 2. 创建会话通道,生产者和mq服务所有通信都在channel通道中完成
            channel = connection.createChannel();
            /**
             * 3. 声明队列
             * 参数：String queue, boolean durable, boolean exclusive, boolean autoDelete, Map<String, Object> arguments
             * 参数明细
             * 1、queue 队列名称
             * 2、durable 是否持久化，如果持久化，mq重启后队列还在
             * 3、exclusive 是否独占连接，队列只允许在该连接中访问，如果connection连接关闭队列则自动删除,如果将此参数设置true可用于临时队列的创建
             * 4、autoDelete 自动删除，队列不再使用时是否自动删除此队列，如果将此参数和exclusive参数设置为true就可以实现临时队列（队列不用了就自动删除）
             * 5、arguments 参数，可以设置一个队列的扩展参数，比如：可设置存活时间
             */
            channel.queueDeclare(queueName, false, false, false, null);

            // 4. 实现消费方法
            commonReceive(channel, queueName, handler, isAutoAck);
        } catch (Exception e) {
            log.error("MQ 消费者启动失败 :{}", e);
        }
    }

    /**
     * 接收处理消息
     *
     * @param exchangeName 消息队列topic
     * @param handler      消息处理类
     * @param isAutoAck    是否自动ACK
     * @throws Exception
     */
    public static void receiveMessageBus(String exchangeName, MessageHandler handler, boolean isAutoAck) throws Exception {
        Connection connection = null;
        Channel channel = null;
        try {
            // 1.获取到连接
            connection = MqUtil.getInstance().newConnection();
            // 2. 创建会话通道,生产者和mq服务所有通信都在channel通道中完成
            channel = connection.createChannel();
            String queueName = UUID.randomUUID().toString();
            // 3、 创建 exchange
            exchangeDeclare(exchangeName);
            /**
             * 3. 声明队列
             * 参数：String queue, boolean durable, boolean exclusive, boolean autoDelete, Map<String, Object> arguments
             * 参数明细
             * 1、queue 队列名称
             * 2、durable 是否持久化，如果持久化，mq重启后队列还在
             * 3、exclusive 是否独占连接，队列只允许在该连接中访问，如果connection连接关闭队列则自动删除,如果将此参数设置true可用于临时队列的创建
             * 4、autoDelete 自动删除，队列不再使用时是否自动删除此队列，如果将此参数和exclusive参数设置为true就可以实现临时队列（队列不用了就自动删除）
             * 5、arguments 参数，可以设置一个队列的扩展参数，比如：可设置存活时间
             */
            channel.queueDeclare(queueName, false, false, false, null);
            // 队列绑定到交换机，key，固定匹配
            channel.queueBind(queueName, exchangeName, exchangeName + ".#");

            // 4. 实现消费方法
            commonReceive(channel, queueName, handler, isAutoAck);
        } catch (Exception e) {
            log.error("MQ 消费者启动失败 {}", exchangeName, e);
        }
    }

    private static void commonReceive(Channel channel, String queueName, MessageHandler handler, boolean isAutoAck) throws IOException {
        // 获取消息，并且处理，这个方法类似事件监听，如果有消息的时候，会被自动调用
        DefaultConsumer consumer = new DefaultConsumer(channel) {
            /**
             * 当接收到消息后此方法将被调用
             *
             * @param consumerTag 消费者标签，用来标识消费者的，在监听队列时设置channel.basicConsume
             * @param envelope    信封，通过envelope
             * @param properties  消息属性
             * @param body        消息内容
             * @throws IOException
             */
            @Override
            public void handleDelivery(String consumerTag, Envelope envelope, AMQP.BasicProperties properties, byte[] body) throws IOException {
                //交换机
                String exchange = envelope.getExchange();
                //消息id，mq在channel中用来标识消息的id，可用于确认消息已接收
                long deliveryTag = envelope.getDeliveryTag();
                // body 即消息体
                String message = new String(body, StandardCharsets.UTF_8);
                try {
                    handler.handle(this.getChannel(), message, envelope, isAutoAck);
                } catch (RejectedExecutionException e) {
                    if (!isAutoAck) {
                        channel.basicAck(envelope.getDeliveryTag(), false);
                    }
                    log.info("Mq message handle queue full");
                } catch (Exception e) {
                    log.error("mq consume message error", e);
                }

            }

            @Override
            public void handleShutdownSignal(String consumerTag, ShutdownSignalException sig) {
                log.error("mq consumer:" + consumerTag + "被关闭", sig);
            }
        };
        /**
         * 监听队列，第二个参数：是否自动进行消息确认。
         * 参数明细：
         * 1、queue 队列名称
         * 2、autoAck 自动回复，当消费者接收到消息后要告诉mq消息已接收，如果将此参数设置为tru表示会自动回复mq，如果设置为false要通过编程实现回复
         * 3、callback，消费方法，当消费者接收到消息要执行的方法
         */
        channel.basicConsume(queueName, isAutoAck, consumer);
    }

    public static void receiveFanoutMessage(String exchangeName, MessageHandler handler, boolean isAutoAck) throws Exception {
        Connection connection = null;
        Channel channel = null;
        try {
            connection = MqUtil.getInstance().newConnection();
            channel = connection.createChannel();
            String queueName = UUID.randomUUID().toString();
            // 3、 创建 exchange
            exchangeDeclare(exchangeName, "fanout");
            channel.queueDeclare(queueName, false, false, false, null);
            channel.queueBind(queueName, exchangeName, exchangeName + ".#");
            // 4. 实现消费方法
            commonReceive(channel, queueName, handler, isAutoAck);
        } catch (Exception e) {
            log.error("MQ 消费者启动失败 {}", exchangeName, e);
        }
    }

    public static void close(Connection connection, Channel channel) {
        if (channel != null) {
            try {
                channel.close();
            } catch (Exception e) {
                log.error("关闭channel失败", e.getMessage());
            }
        }
        if (connection != null) {
            try {
                connection.close();
            } catch (Exception e) {
                log.error("关闭connection失败", e.getMessage());
            }
        }
    }

    /**
     * 定义消息处理类接口
     */
    public static interface MessageHandler {
        void handle(Channel channel, String message, Envelope envelope, boolean autoAck);
    }

    public static void main(String[] args) throws Exception {
        // 4、消息内容
        String message = "Hello World!" + System.currentTimeMillis();
//        MqUtil.sendMessage("receive-test-20", "666", message);
        //MqUtil.receiveMessage();

    }

}
