package net.wanji.common.tool.resttool;

import com.alibaba.fastjson.JSON;
import net.wanji.common.utils.tool.StringUtils;
import lombok.Getter;
import org.springframework.http.*;
import org.springframework.web.client.RestTemplate;

import java.util.Map;

import static org.apache.http.HttpStatus.SC_OK;

/**
 * @author wanji
 * todo spring框架默认处理MimeType方式过于严格，存在兼容性问题，需要使用原生重新封装
 * todo 不支持的接口示例：    @RequestMapping(value = "/zmove_material")
 *     public void save(@QueryParam("sap-client") String sap_client, @RequestBody User entity, HttpServletResponse response) {
 *         User user = User.newInstance();
 *         user.setName("name2");
 *         user.setId("id_1");
 *         user.setUsername(sap_client);
 *         response.setContentType("TEXT;charset=utf-8");
 *         try(OutputStream ros = response.getOutputStream()) {
 *             IOUtils.write("xxx", ros);
 *             ros.flush();
 *         } catch (IOException e) {
 *         }
 *     }
 * todo 参见 org.springframework.util.MimeTypeUtils#parseMimeTypeInternal(java.lang.String)
 *
 */
public class RestTemplateTool {

    @Getter
    private static RestTemplate restTemplate = RestTemplateConfiguration.newInstance().getRestTemplate();

    private RestTemplateTool() {
    }

    /**
     * get
     *
     * @param url
     * @return
     */
    public static String get(String url) throws Exception {
        if (StringUtils.isBlank(url)) {
            return null;
        }
        ResponseEntity<String> responseEntity = restTemplate.getForEntity(url, String.class);
        if (200 != responseEntity.getStatusCode().value()) {
            throw new Exception(responseEntity.toString());
        }
        return responseEntity.getBody();
    }

    /**
     * delete
     *
     * @param url
     * @return
     */
    public static void delete(String url) {
        if (!StringUtils.isNotBlank(url)) {
            return;
        }
        restTemplate.delete(url, String.class);
    }

    /**
     * post
     *
     * @param url
     * @param requestEntity
     * @return
     */
    public static String post(String url, HttpEntity requestEntity) throws Exception {
        if (StringUtils.isBlank(url)) {
            return null;
        }
        ResponseEntity<String> responseEntity = restTemplate.postForEntity(url, requestEntity, String.class);
        if (200 != responseEntity.getStatusCode().value()) {
            throw new Exception(responseEntity.toString());
        }
        return responseEntity.getBody();
    }

    /**
     * post
     *
     * @param url
     * @param param
     * @return
     */
    public static String post(String url, String param) throws Exception {
        HttpHeaders requestHeaders = new HttpHeaders();
        requestHeaders.setContentType(MediaType.APPLICATION_JSON_UTF8);
        HttpEntity requestEntity = new HttpEntity<>(JSON.parseObject(param), requestHeaders);
        return post(url, requestEntity);
    }

    /**
     * post
     *
     * @param url
     * @param param
     * @param requestHeaders
     * @return
     */
    public static String post(String url, String param, HttpHeaders requestHeaders) throws Exception {
        HttpEntity<Map<String, Object>> requestEntity = new HttpEntity<>(JSON.parseObject(param), requestHeaders);
        return post(url, requestEntity);
    }

    /**
     * post
     *
     * @param url
     * @param param
     * @param headerStr
     * @return
     */
    public static String post(String url, String param, String headerStr) throws Exception {
        HttpHeaders requestHeaders = new HttpHeaders();
        String[] headers = headerStr.split("\n");
        for (String h : headers) {
            String[] headerEntry = h.split(":");
            requestHeaders.add(headerEntry[0], headerEntry[1]);
        }
        return post(url, param, requestHeaders);
    }

    /**
     * put
     *
     * @param url
     * @param requestEntity
     */
    public static void put(String url, HttpEntity<String> requestEntity) {
        if (!StringUtils.isNotBlank(url)) {
            return;
        }
        restTemplate.put(url, requestEntity, String.class);
    }

    /**
     * put
     *
     * @param url
     * @param param
     */
    public static void put(String url, String param) {
        HttpHeaders requestHeaders = new HttpHeaders();
        requestHeaders.setContentType(MediaType.APPLICATION_JSON_UTF8);
        HttpEntity<String> requestEntity = new HttpEntity<>(param, requestHeaders);
        put(url, requestEntity);
    }

    /**
     * put
     *
     * @param url
     * @param param
     * @param requestHeaders
     */
    public static void put(String url, String param, HttpHeaders requestHeaders) {
        HttpEntity<String> requestEntity = new HttpEntity<>(param, requestHeaders);
        put(url, requestEntity);
    }

    /**
     * put
     *
     * @param url
     * @param param
     * @param headerStr
     */
    public static void put(String url, String param, String headerStr) {
        HttpHeaders requestHeaders = new HttpHeaders();
        String[] headers = headerStr.split("\n");
        for (String h : headers) {
            String[] headerEntry = h.split(":");
            requestHeaders.add(headerEntry[0], headerEntry[1]);
        }
        put(url, param, requestHeaders);
    }

    /**
     * get请求，可添加header和request params。
     * <p>
     * url类型：http://127.0.0.1:15020/deleting?ids={ids}
     *
     * @param url     request地址。
     * @param headers 请求头。
     * @param params  请求参数。
     * @return 响应对象。
     * @throws Exception 如果请求异常。
     */
    public static ResponseEntity get(String url, Map<String, Object> headers, Map<String, Object> params) throws Exception {
        // 添加header
        HttpHeaders httpHeaders = new HttpHeaders();
        for (Map.Entry<String, Object> header : headers.entrySet()) {
            httpHeaders.add(header.getKey(), (String) header.getValue());
        }
        //HttpEntity
        HttpEntity<Object> httpEntity = new HttpEntity<>(params, httpHeaders);
        ResponseEntity<?> responseEntity = null;
        if (params == null) {
            responseEntity = restTemplate.exchange(url, HttpMethod.GET, httpEntity, String.class);
        } else {
            responseEntity = restTemplate.exchange(url, HttpMethod.GET, httpEntity, String.class, params);
        }

        if (SC_OK != responseEntity.getStatusCode().value()) {
            throw new Exception(responseEntity.toString());
        }
        return responseEntity;
    }

    /**
     * post请求，可添加header和json字符串格式的body。
     *
     * @param url     请求路径。
     * @param headers 请求头。
     * @param body    请求体。
     * @return 响应参数。
     * @throws Exception 如果请求异常。
     */
    public static ResponseEntity post(String url, Map<String, Object> headers, Map<String, Object> body) throws Exception {
        // 添加header
        HttpHeaders httpHeaders = new HttpHeaders();
        if (!headers.containsKey("Content-Type")) {
            httpHeaders.setContentType(MediaType.APPLICATION_JSON);
        }
        if (!headers.containsKey("Accept")) {
            httpHeaders.add("Accept", "*/*");
        }
        //设置返回媒体数据类型
        if (!headers.isEmpty()) {
            for (Map.Entry<String, Object> header : headers.entrySet()) {
                httpHeaders.add(header.getKey(), (String) header.getValue());
            }
        }
        HttpEntity<Object> request = new HttpEntity<>(body, httpHeaders);
        ResponseEntity<String> responseEntity;
        if (body == null || body.isEmpty()) {
            responseEntity = restTemplate.postForEntity(url, request, String.class);
        } else {
            responseEntity = restTemplate.postForEntity(url, request, String.class, body);
        }
        if (SC_OK != responseEntity.getStatusCode().value()) {
            throw new Exception(responseEntity.toString());
        }
        return responseEntity;
    }

}
